#include <grammar.h>
#include <first.h>
#ifdef DEBUG
#include <stdio.h>
#endif

typedef struct
{
  symbol_t* left;
  rule_t* right;
  GList* dot;
  symbol_t* lookahead;
} item_t;

item_t* item_new (symbol_t* left, rule_t* right, symbol_t* lookahead)
{
  item_t* item;
  item = g_malloc (sizeof (item_t));
  item->left = left;
  item->right = right;
  item->dot = grammar_get_rule (right);
  item->lookahead = lookahead;
  return item;
}

item_t* item_copy (item_t* item)
{
  item_t* newitem;
  int n;
  newitem = item_new (symbol_copy (item->left), rule_copy (item->right),
		      symbol_copy (item->lookahead));
  n = g_list_position (grammar_get_rule (item->right), item->dot);
  newitem->dot = g_list_nth (grammar_get_rule (newitem->right), n);
  return newitem;
}

gint item_cmp (const item_t* a, const item_t* b)
{
  int c;
  int na;
  int nb;
  if ((c = symbol_cmp (a->left, b->left)) != 0)
    return c;
  if ((c = rule_cmp (a->right, b->right)) != 0)
    return c;
  if ((c = symbol_cmp (a->lookahead, b->lookahead)) != 0)
    return c;
  na = g_list_position (grammar_get_rule (a->right), a->dot);
  nb = g_list_position (grammar_get_rule (b->right), b->dot);
  if (na < nb)
    return -1;
  else if (na > nb)
    return 1;
  return 0;
}

gboolean item_equal (gconstpointer data1, gconstpointer data2)
{
  return (item_cmp (data1, data2) == 0);
}

guint item_hash (gconstpointer data)
{
  item_t* item;
  guint hash;
  item = (item_t*) data;
  hash = rule_hash (item->right) * 37 + symbol_hash (item->left);
  hash = hash * 37 + symbol_hash (item->lookahead);
  return hash;
}

void item_delete (item_t* item)
{
  g_free (item->left);
  rule_delete (item->right);
  g_free (item->lookahead);
  g_free (item);
}

#ifdef DEBUG
void item_print (item_t* item)
{
  GList* l;
  fprintf (stdout, "%s -> ", g_quark_to_string (item->left->value));
  l = grammar_get_rule (item->right);
  while (l != NULL)
    {
      symbol_t* symbol;
      symbol = (symbol_t*) l->data;
      if (l == item->dot)
	{
	  fprintf (stdout, ".");
	}
      fprintf (stdout, " %s ", g_quark_to_string (symbol->value));
      l = g_list_next (l);
    }
  if (item->dot == NULL)
    {
      fprintf (stdout, ".");
    }
  fprintf (stdout, ", %s\n", g_quark_to_string (item->lookahead->value));
}
#endif

GHashTable* item_set_new ()
{
  return g_hash_table_new_full (item_hash, item_equal, item_delete, NULL);
}

gboolean item_set_add (GHashTable* item_set, item_t* item)
{
  if (!g_hash_table_lookup_extended (item_set, item, NULL, NULL))
    {
      g_hash_table_insert (item_set, item, NULL);
      return TRUE;
    }
  return FALSE;
}

static void put_key_on_list (gpointer key, gpointer val, gpointer data)
{
  GList** l;
  l = (GList**) data;
  *l = g_list_prepend (*l, key);
}

gboolean item_set_equal (gconstpointer a, gconstpointer b)
{
  GList* l;
  gboolean equal;
  if (g_hash_table_size (a) != g_hash_table_size (b))
    return FALSE;
  l = NULL;
  g_hash_table_foreach (a, put_key_on_list, (gpointer) &l);
  equal = TRUE;
  while (l != NULL && equal == TRUE)
    {
      equal = g_hash_table_lookup_extended (b, l->data, NULL, NULL);
      l = g_list_next (l);
    }
  g_list_free (l);
  return equal;
}

void hash_item (gpointer key, gpointer val, gpointer data)
{
  guint* hash;
  hash = (guint*) data;
  *hash = *hash * 37 + item_hash (key);
}

guint item_set_hash (gconstpointer a)
{
  guint hash = 0;
  g_hash_table_foreach (a, hash_item, &hash);
  return hash;
}

void item_set_add_each (gpointer key, gpointer val, gpointer data)
{
  item_set_add (data, item_copy (key));
}

GHashTable* item_set_copy (GHashTable* item_set)
{
  GHashTable* newitem_set;
  newitem_set = item_set_new ();
  g_hash_table_foreach (item_set, item_set_add_each, newitem_set);
  return newitem_set;
}

#ifdef DEBUG
void item_set_print_each (gpointer key, gpointer val, gpointer data)
{
  item_print (key);
}

void item_set_print (GHashTable* item_set)
{
  g_hash_table_foreach (item_set, item_set_print_each, NULL);
}
#endif

rule_t* rule_new_item (item_t* item)
{

  rule_t* rule;
  GList* l;
  rule = rule_new ();
  l = g_list_next (item->dot);
  while (l != NULL)
    {
      rule_append (rule, symbol_copy (l->data));
      l = g_list_next (l);
    }
  rule_append (rule, symbol_copy (item->lookahead));
  return rule;

}

void item_set_closure_step (GHashTable* item_set, grammar_t* grammar,
			    GHashTable* first, item_t* item)
{
  if (item->dot != NULL)
    {
      symbol_t* symbol;
      symbol = (symbol_t*) item->dot->data;
      if (symbol->terminal == FALSE)
	{
	  GList* rules;
	  GList* terminals;
	  rule_t* rule;
	  rule = rule_new_item (item);
	  terminals = first_rule (first, rule);
	  rule_delete (rule);
	  rules = grammar_get_rules (grammar, symbol);
	  while (rules != NULL)
	    {
	      GList* lookahead;
	      lookahead = terminals;
	      while (lookahead != NULL)
		{
		  rule_t* rule;
		  item_t* newitem;
		  rule = rule_copy (rules->data);
		  newitem = item_new (symbol_copy (symbol), rule,
				      symbol_copy (lookahead->data));
		  if (!item_set_add (item_set, newitem))
		    item_delete (newitem);
		  lookahead = g_list_next (lookahead);
		}
	      rules = g_list_next (rules);
	    }
	  g_list_free (terminals);
	}
    }
}

GHashTable* item_set_closure (GHashTable* item_set, grammar_t* grammar,
			      GHashTable* first)
{
  int size;
  int last_size;
  GList* l;
  size = g_hash_table_size (item_set);
  do
    {
      last_size = size;
      l = NULL;
      g_hash_table_foreach (item_set, put_key_on_list, &l);
      while (l != NULL)
	{
	  item_set_closure_step (item_set, grammar, first, l->data);
	  l = g_list_next (l);
	}
      g_list_free (l);
      size = g_hash_table_size (item_set);
    } while (last_size < size);
  return item_set;
}

GHashTable* item_set_goto (GHashTable* item_set, grammar_t* grammar,
			   GHashTable* first, symbol_t* symbol)
{
  GList* l;
  GHashTable* newitem_set;
  newitem_set = item_set_new ();
  l = NULL;
  g_hash_table_foreach (item_set, put_key_on_list, &l);
  while (l != NULL)
    {
      item_t* item;
      item = item_copy (l->data);
      if (item->dot == NULL || !symbol_equal (item->dot->data, symbol))
	{
	  item_delete (item);
	}
      else
	{
	  item->dot = g_list_next (item->dot);
	  if (!item_set_add (newitem_set, item))
	    item_delete (item);
	}
      l = g_list_next (l);
    }
  return item_set_closure (newitem_set, grammar, first);
}


/*
 * This part is a little tricky. We can simply take an approach similar
 * to the goto approach, adding a new set for every other set and grammar
 * symbol every time we get a new set.
 * Although a similar argument about the goto may apply (which is a very
 * good reason to optimize it), the case for the collection is worse, since
 * we will be building a new set for the number of sets in the collection
 * times the number of symbol grammar. (Not building in fact, but computing
 * it).
 * Here is an approach that may work:
 * For each *new* item set, we compute its gotos. Each goto is looked up
 * in the collection. If it's a new item set, it's included in a list of
 * new item sets. To compute the gotos of an item set, we should pick only
 * the grammar symbols that appear in the items' dots. That solves the
 * problem of getting a list of all the grammar symbols.
 * So, there will be a list of item sets that we must iterate through. We
 * add each new list to the hash table too, so we can look it up when we
 * get a new one. When the list is finished, we are done and do not have to
 * check the hash table size for each iteration.
 * So, we need the following functions:
 * One to check if an item set is already there and add it if it's not.
 * One to get the symbols hash table, so we can put the computed goto
 * there anyway. So, that's an add and a lookup.
 */

/*
 * TODO: associate a number to each item set.
 * Use a structure to the collection with a counter.
 * Use another structure to the value in the hash table for each item set.
 * This structure will have the symbol hash table (for the goto) and the
 * number associated with the item set.
 * In fact, the counter may be the hash table size.
 */

GHashTable* item_collection_new ()
{
  return g_hash_table_new_full (item_set_hash, item_set_equal,
				g_hash_table_destroy, g_hash_table_destroy);
}

gboolean item_collection_add (GHashTable* collection, GHashTable* item_set,
			      GHashTable** key)
{
  if (!g_hash_table_lookup_extended (collection, item_set,
				     (gpointer*)key, NULL))
    {
      GHashTable* symbols;
      symbols = g_hash_table_new_full (symbol_hash, symbol_equal,
				       g_free, NULL);
      g_hash_table_insert (collection, item_set, symbols);
      return TRUE;
    }
  return FALSE;
}

GHashTable* item_collection_lookup (GHashTable* collection,
				    GHashTable* item_set)
{
  GHashTable* symbols;
  if (!g_hash_table_lookup_extended (collection, item_set,
				     NULL, (gpointer*)&symbols))
    {
      return NULL;
    }
  return symbols;
}

#define HASH_ITEM_SET(item_set) (((GPOINTER_TO_INT(item_set) & 0x3f00) >> 8))
#ifdef DEBUG
void item_collection_print_each (gpointer key, gpointer val, gpointer data)
{
  GHashTable* item_set;
  item_set = (GHashTable*) key;
  fprintf (stdout, "Item %x:\n", HASH_ITEM_SET(key));
  item_set_print (item_set);
  fprintf (stdout, "\n");
}

void item_set_print_goto (gpointer key, gpointer val, gpointer data)
{
  symbol_t* symbol;
  symbol = (symbol_t*) key;
  fprintf (stdout, "GOTO (%x, %s) =\t %x\n", HASH_ITEM_SET(data),
	   g_quark_to_string (symbol->value), HASH_ITEM_SET(val));
}

void item_collection_print_goto (gpointer key, gpointer val, gpointer data)
{
  GHashTable* symbols;
  symbols = (GHashTable*) val;
  g_hash_table_foreach (symbols, item_set_print_goto, key);
  fprintf (stdout, "\n");
}

void item_collection_print (GHashTable* collection)
{
  g_hash_table_foreach (collection, item_collection_print_each, NULL);
  g_hash_table_foreach (collection, item_collection_print_goto, NULL);
}
#endif

GHashTable* item_collection_goto (GHashTable* collection, grammar_t* grammar,
				  GHashTable* first, GHashTable* item_set,
				  symbol_t* symbol)
{
  GHashTable* symbols;
  GHashTable* newitem_set;
  GHashTable* goto_item_set;
  GHashTable* old_item_set;
  newitem_set = item_set_copy (item_set);
  if (!item_collection_add (collection, newitem_set, NULL))
    {
      g_hash_table_destroy (newitem_set);
    }
  symbols = item_collection_lookup (collection, item_set);
  if (g_hash_table_lookup_extended (symbols, symbol, NULL, NULL))
    {
      return NULL;
    }
  goto_item_set = item_set_goto (item_set, grammar, first, symbol);
  if (!item_collection_add (collection, goto_item_set, &old_item_set))
    {
      g_hash_table_insert (symbols, symbol, old_item_set);
      g_hash_table_destroy (goto_item_set);
      return NULL;
    }
  else
    {
      g_hash_table_insert (symbols, symbol, goto_item_set);
      return goto_item_set;
    }
}

void item_set_collection (grammar_t* grammar, GHashTable* first, symbol_t* start)
{
  GHashTable* collection;
  GHashTable* item_set;
  item_t* item;
  rule_t* rule;
  GList* new_item_sets;
  rule = rule_new ();
  rule_append (rule, symbol_copy (start));
  item = item_new (symbol_new (FALSE, -1), rule, symbol_new (TRUE, 0));
  item_set = item_set_new ();
  item_set_add (item_set, item);
  item_set_closure (item_set, grammar, first);
  collection = g_hash_table_new_full (item_set_hash, item_set_equal,
				      g_hash_table_destroy, NULL);
  item_collection_add (collection, item_set, NULL);
  new_item_sets = g_list_append (NULL, item_set);
  while (new_item_sets != NULL)
    {
      GHashTable* next_item_set;
      GHashTable* new_item_set;
      GList* items;
      next_item_set = (GHashTable*) new_item_sets->data;
      items = NULL;
      g_hash_table_foreach (next_item_set, put_key_on_list, &items);
      while (items != NULL)
	{
	  item_t* item;
	  symbol_t* symbol;
	  item = (item_t*) items->data;
	  if (item->dot != NULL)
	    {
	      symbol = (symbol_t*) item->dot->data;
	      if ((new_item_set =
		   item_collection_goto (collection, grammar, first,
					 next_item_set, symbol)) != NULL)
	      {
		g_list_append (new_item_sets, new_item_set);
	      }
	    }
	  items = g_list_next (items);
	}
      g_list_free (items);
      new_item_sets = g_list_next (new_item_sets);
    }
  g_list_free (new_item_sets);
#ifdef DEBUG
  item_collection_print (collection);
#endif
  g_hash_table_destroy (collection);
}
